window.Gleam.initSearch([{"doc":"glam","title":"glam","content":"# glam\n\n[![Package Version](https://img.shields.io/hexpm/v/glam)](https://hex.pm/packages/glam)\n[![Hex Docs](https://img.shields.io/badge/hex-docs-ffaff3)](https://hexdocs.pm/glam/)\n![CI](https://github.com/giacomocavalieri/glam/workflows/CI/badge.svg?branch=main)\n\nA package to help you easily pretty print structured data ✨\n\n&gt; ⚙️ This package works with both the Erlang and Javascript target\n\n## Installation\n\nTo add this package to your Gleam project:\n\n```sh\ngleam add glam\n```\n\n## Pretty printing\n\nWhen working with structured data (like JSON, XML, lists, ASTs, ...) it can\nbe useful to print it in a nice and tidy way.\nOf course, one could go the quick and dirty way and just `string.inspect`\nthe data structure and call it a day.\nHowever, the result would hardly be readable for complex data structures.\n\nOn the other hand, hand writing a pretty printing function can be quite\ndifficult (I sure have spent my fair share of time writing those) and error\nprone: there&#39;s a lot of moving pieces to juggle and its easy for bugs to sneak\nin.\n\nA pretty printing package like glam provides some basic building block to\n_describe_ the structure of your data and takes care of the heavy lifting of\nactually finding the best layout to format the data, gracefully handling line\nlimits and indentation.\n\n## Getting started\n\nIf you want to get started with glam, a great starting point is the\n[introductory tutorial](https://hexdocs.pm/glam/01_gleam_lists.html) which will guide\nyou step by step in writing a pretty printer for lists.\n\nBy the end of the tutorial you&#39;ll know most of the package&#39;s API and you&#39;ll have\nimplemented a function that can nicely display lists like this:\n\n```gleam\npretty_list([&quot;Gleam&quot;, &quot;is&quot;, &quot;fun!&quot;])\n|&gt; doc.to_string(10) // Fit the list on lines of width 10\n\n// [\n//   &quot;Gleam&quot;,\n//   &quot;is&quot;,   \n//   &quot;fun!&quot;,   \n// ]\n```\n\nYou can also have a look at the\n[`examples`](https://github.com/giacomocavalieri/glam/tree/main/src/examples)\nfolder in the GitHub repo, there you&#39;ll also find a JSON pretty printer.\n\n## Contributing\n\nIf you think there&#39;s any way to improve this package, or if you spot a bug\ndon&#39;t be afraid to open PRs, issues or requests of any kind!\nAny contribution is welcome 💜\n","url":"index.html"},{"doc":"glam","title":"glam","content":"# Pretty printing glamorous lists ✨\n\n## Goals\n\nThe goal of this tutorial is to get you started on pretty printing with the\nglam package introducing its API a step at a time.\n\nAs a running example we&#39;re going to write a pretty printer for lists of strings.\nThere are three rules we want our pretty printer to follow:\n\n- If a list is shorter than the maximum allowed line width, it is kept on\n  a single line\n- If it is longer than the maximum line width, each element is placed on a new\n  line and is indented by two spaces; in addition the last element has a\n  trailing comma\n- All strings are surrounded by double quotes `&quot;`\n\nSince a ~~picture~~ code snippet is worth a thousand words let me show you an\nexample:\n\n```gleam\n// ----------------- // &lt;- This is to show the 20 chars limit\n\n[&quot;Hello&quot;, &quot;Gleam&quot;]   // &lt;- This can stay on a single line, easy!\n\n[                    // &lt;- This wouldn&#39;t fit on a single line\n  &quot;Gleam&quot;,           //    so each element is split on a different   \n  &quot;is&quot;,              //    line and is indented by two spaces \n  &quot;fun!&quot;,            // &lt;- Notice the trailing comma!\n]\n```\n\nNow that we have a clear picture of what a pretty printed lists should look\nlike we can dive deep into the wonders of pretty printing!\n\n## Setup\n\nThis tutorial is full of code examples so you can follow along with your own\nGleam project. If you have a working installation of Gleam, open your terminal\nand enter the following:\n\n```shell\ngleam new glam_tutorial # create a new project named `glam_tutorial`\ncd glam_tutorial        # jump into its directory\ngleam add glam          # add the `glam` package to the project&#39;s dependencies\n```\n\n&gt; If you know nothing about Gleam and somehow ended up here, I can recommend you\n&gt; to first check out the\n&gt; [Gleam&#39;s Exercism track]((https://exercism.org/tracks/gleam/concepts)).\n&gt; It&#39;s a great resource to get started!\n\nFirst things first, you can import the glam package in your project&#39;s\nmodules like this:\n\n```gleam\nimport glam/doc.{Document}\n```\n\n## Basic building blocks\n\nLet&#39;s start by writing the type of the function that can turn lists into pretty\nprinted lists; after all it is always a good idea to write down the types and\nlet them guide you:\n\n```gleam\npub fn pretty_list(list: List(String)) -&gt; Document {\n  todo(&quot;Write an amazing pretty printer&quot;)\n}\n```\n\nThe first thing you can notice is that the pretty printing function&#39;s return\ntype is `Document`. But wait, weren&#39;t we supposed to pretty print the list?\nWhy isn&#39;t the function returning a `String`?\n\nHere is a crucial point of how the glam package handles pretty printing:\nyou can _describe_ the structure of the text to pretty print with some basic\nbuilding blocks, and glam will do the heavy lifting of trying to find the\nbest layout for you.\nAs you may have guessed those building blocks are `Document`s.\n\nAs long as you can turn something into a `Document`, glam will pretty print it\nfor you. Quite neat, isn&#39;t it?\n\n### A naive approach\n\nFirst of all, we need a way to create a `Document` from a `String`,\nluckily glam has got us covered with the `doc.from_string` function which does\nexactly that:\n\n```gleam\ndoc.from_string(&quot;Hello, world!&quot;)\n|&gt; doc.to_string(80)\n// -&gt; Hello, world!\n```\n\nThere are a few things to unpack in this example:\n\n- `doc.from_string` takes a string and turns it into a `Document`\n- `doc.to_string` does the opposite: given a maximum line width and a document,\n  it turns it into a pretty printed string. Later we will discover how the\n  maximum line width can influence the pretty printing process\n\nWith `doc.from_string` in our toolbox we can now turn each of the items of the\nlist into documents:\n\n```gleam\npub fn pretty_list(list: List(String)) -&gt; Document {\n  let list_item_to_document = fn(item) { doc.from_string(&quot;\\&quot;&quot; &lt;&gt; item &lt;&gt; &quot;\\&quot;&quot;) }\n  let docs = list.map(list, list_item_to_document)\n\n  todo(&quot;Still not enough...&quot;)\n}\n```\n\n`docs` is a list of documents but in order to be printed with `doc.to_string`\nit needs to be turned into a single `Document`.\nTo do that we can use the `doc.concat` funtion which takes a list of documents\nas input and joins them together into a single document:\n\n```gleam\npub fn pretty_list(list: List(String)) -&gt; Document {\n  let list_item_to_document = fn(item) { doc.from_string(item) }\n  let docs = list.map(list, list_item_to_document)\n  docs |&gt; doc.concat\n}\n```\n\nWe&#39;re now ready to try `pretty_list`:\n\n```gleam\n[&quot;Hello&quot;, &quot;world!&quot;]\n|&gt; pretty_list\n|&gt; doc.to_string(80)\n// -&gt; &quot;Hello&quot;&quot;world!&quot;\n```\n\nNot quite list-looking but we&#39;re getting there... adding commas and brackets can\nbe done with some standard library functions on lists:\n\n```gleam\npub fn pretty_list(list: List(String)) -&gt; Document {\n  let list_item_to_document = fn(item) { doc.from_string(&quot;\\&quot;&quot; &lt;&gt; item &lt;&gt; &quot;\\&quot;&quot;) }\n\n  list.map(list, list_item_to_document)\n  |&gt; list.intersperse(doc.from_string(&quot;, &quot;)) // put a comma between each doc\n  |&gt; list.prepend(doc.from_string(&quot;[&quot;))      // starting open bracket\n  |&gt; list.append([doc.from_string(&quot;]&quot;)])     // final closing bracket\n  |&gt; doc.concat                              // join everything together\n}\n```\n\n```gleam\n[&quot;Hello&quot;, &quot;world!&quot;]\n|&gt; pretty_list\n|&gt; doc.to_string(80)\n// -&gt; [&quot;Hello&quot;, &quot;world!&quot;]\n```\n\nAwesome, with a couple of lines of code you can now pretty print a list of\nstrings.\nLet&#39;s play a bit more with the maximum line width and see how the list gets\nprinted:\n\n```gleam\npub fn main() {\n  let list = [&quot;Hello&quot;, &quot;world!&quot;]\n  pretty_list(list)\n  |&gt; doc.to_string(5) // &lt;- this is really small and the list\n  |&gt; io.println       //    should be printed on multiple lines\n}\n\n// main()\n// -&gt; [&quot;Hello&quot;, &quot;world!&quot;]\n```\n\nWe would expect the list to be split on multiple lines like in the example I\nshowed you at the beginning of the tutorial.\nIt looks like `doc.concat` is not enough to create beautiful documents. In the\nnext section we&#39;ll discover more building blocks that can get us closer to a\npretty printed list.\n\n### A less naive approach\n\nTo understand the problem here we need to get into a bit more detail on how the\npretty printing of documents works:\n\n- when the pretty printer has to display a `doc.from_string` it simply outputs\n  the string passed to it as an argument leaving it unchanged\n- when the pretty printer has to display a list of documents joined with\n  `doc.concat`, it formats each document one and joins them together without\n  adding any whitespace or trying to split them to fit on a line\n\n&gt; If you already have some experience in Gleam and functional programming\n&gt; you may have noticed that a pretty printer that only has these two primitives\n&gt; is just a more convoluted way of concatenating strings, `doc.concat` would act\n&gt; as a `list.fold(strings, from: &quot;&quot;, with: fn(acc, string) { acc &lt;&gt; string })`\n&gt;\n&gt; If this `fold` thing looks like gibberish to you, don&#39;t worry!\n&gt; You can look at the\n&gt; [Gleam&#39;s Exercism track](https://exercism.org/tracks/gleam/concepts), it&#39;s a\n&gt; great way to get started and hone your functional programming skills! (At\n&gt; this point I&#39;m sure you get how much I love Gleam&#39;s Exercism track)\n\n#### Grouping docs\n\nWhat we really need is a way to explicitly tell the pretty printer it is\nallowed to break a document if it does not fit on a single line.\n\nThis is the job for `doc.group` and `doc.space`:\n\n- when the pretty printer runs into a `doc.group` it first tries to render the\n  document on a single line, just like `doc.concat`. The catch is that if the\n  document would not fit on a single line, the pretty printer will break all the\n  spaces inside it\n- a `doc.space` is a special kind of document that gets normally rendered as a\n  single whitespace: `&quot; &quot;`. However, the pretty printer can turn these spaces\n  into newlines in order to split a `doc.group` that would not fit on a single\n  line\n\nLet&#39;s look at an example to make things clearer:\n\n```gleam\nlet doc =\n  [doc.from_string(&quot;Hello&quot;), doc.space, doc.from_string(&quot;world!&quot;)]\n  |&gt; doc.concat\n  |&gt; doc.group\n\ndoc.to_string(doc, 80)\n// -&gt; Hello world!\ndoc.to_string(doc, 10)\n// -&gt;\n// Hello\n// world!\n```\n\nAs you can see, in the second example, where the line width wouldn&#39;t allow\n`&quot;Hello&quot;` and `&quot;world!&quot;` to fit onto a single line, the space is rendered as a\nnewline.\n\n#### Even more examples\n\nBefore going back to our list pretty printer we should experiment a little bit\nmore with `doc.group` and `doc.space`.\n\nIn particular there&#39;s a little detail I omitted in my first explanation of how\ngroup formatting works. Let&#39;s see how the pretty printer deals with nested\ngroups:\n\n```gleam\nlet food =\n  [&quot;lasagna&quot;, &quot;ravioli&quot;, &quot;pizza&quot;]\n  |&gt; list.map(doc.from_string)\n  |&gt; list.intersperse(with: doc.space)\n  |&gt; doc.concat\n  |&gt; doc.group\n\nlet food_i_love =\n  // Sorry for being stereotypically italian\n  [doc.from_string(&quot;Food I love:&quot;), doc.space, food]\n  |&gt; doc.concat\n  |&gt; doc.group\n```\n\n&gt; You&#39;ll find out that we frequently need to do something like this to join\n&gt; documents together using a custom separator:\n&gt;\n&gt; ```gleam\n&gt; list_of_docs\n&gt; |&gt; list.intersperse(with: some_separator)\n&gt; |&gt; doc.concat\n&gt; ```\n&gt;\n&gt; To make things a bit easier to read you can also use the `doc.join` function\n&gt; which is a shorthand for an `intersperse` followed by `concat`:\n&gt;\n&gt; ```gleam\n&gt; list_of_docs\n&gt; |&gt; doc.join(with: some_separator)\n&gt; ```\n\nHere we have two groups, one inside the other, let&#39;s play around with the\nmaximum line width and see how this gets printed:\n\n```gleam\nfood_i_love |&gt; doc.to_string(80)\n// -&gt; Food I love: lasagna ravioli pizza\n\nfood_i_love |&gt; doc.to_string(20)\n// -&gt;\n// Food I love:\n// lasagna\n// ravioli\n// pizza\n\nfood_i_love |&gt; doc.to_string(30)\n// -&gt;\n// Food I love:\n// lasagna ravioli pizza\n```\n\n- In the first example we gave the document enough space to fit on a single\n  line, all spaces get rendered as a normal whitespace and everything ends up on\n  a single line\n- In the second example the line width is quite small: neither the outer, nor\n  the inner document could fit on a single line so all spaces become newlines\n- The most interesting example is the third one where the outer group is split\n  while the inner is not.\n  What end up happening in the pretty process could be described like this:\n  - The pretty printer first tries to render everything on a single line but\n    the resulting document would exceed the maximum width of 30 chars\n  - So it decides to split the outer group and its space gets rendered as a\n    newline\n  - On the new line the pretty printer tries to render the inner document\n    and it fits just right! So it&#39;s spaces, contrary to what happened to those\n    of the outer group, are not turned into newlines\n\nSo when the pretty printer ends up breaking a group it may still not break the\ninner groups it&#39;s composed of. Each subgroup is considered _separately_.\n\n&gt; It can get some time to wrap your head around how groups can influence the\n&gt; pretty printing process.\n&gt; If you want to test your understanding of `doc.group`, try to look at this\n&gt; example and guess how it will get printed at different line widths:\n&gt;\n&gt; ```gleam\n&gt; let abcd =\n&gt;   [&quot;a&quot;, &quot;b&quot;, &quot;c&quot;, &quot;d&quot;]\n&gt;   |&gt; list.map(doc.from_string)\n&gt;   |&gt; doc.join(with: doc.space)\n&gt;   |&gt; doc.group\n&gt;\n&gt; let efgh =\n&gt;   [&quot;e&quot;, &quot;f&quot;, &quot;g&quot;, &quot;h&quot;]\n&gt;   |&gt; list.map(doc.from_string)\n&gt;   |&gt; doc.join(with: doc.space)\n&gt;   |&gt; doc.group\n&gt;\n&gt; let doc =\n&gt;   [doc.from_string(&quot;some letters:&quot;), abcd, efgh]\n&gt;   |&gt; doc.join(with: doc.space)\n&gt;   |&gt; doc.group\n&gt;\n&gt; // Try different widths and guess what the document will look like\n&gt; doc |&gt; doc.to_string(...) \n&gt; ```\n\n#### Back onto lists\n\nIt shouldn&#39;t be too hard to update the previous iteration of `pretty_list` to\ntake advantage of groups and breakable spaces. Let&#39;s try it out:\n\n```gleam\npub fn pretty_list(list: List(String)) -&gt; Document {\n  let list_item_to_document = fn(item) { doc.from_string(&quot;\\&quot;&quot; &lt;&gt; item &lt;&gt; &quot;\\&quot;&quot;) }\n  let comma = doc.concat([doc.from_string(&quot;,&quot;), doc.space])\n  let open_square = doc.concat([doc.from_string(&quot;[&quot;), doc.space])\n  let close_square = doc.concat([doc.space, doc.from_string(&quot;]&quot;)])\n\n  list.map(list, list_item_to_document)\n  |&gt; doc.join(with: comma)\n  |&gt; doc.prepend(open_square)\n  |&gt; doc.append(close_square)\n  |&gt; doc.group\n  // ^-- Don&#39;t forget the group, otherwise the pretty\n  //     printer won&#39;t be able to split on spaces\n}\n```\n\nThe main difference is that the comma that we use as a separator is now followed\nby a `doc.space` which the pretty printer can break if the list is too long and\ndoes now fit on a single line. The same goes for square brackets: the open\nbracket is followed by a `doc.space` so that, if the document gets broken, the\nlist&#39;s first item will end up on a new line. Likewise, the closed bracket is\npreceded by a space so that it can end up on its own line:\n\n```gleam\npretty_list([&quot;Gleam&quot;, &quot;is&quot;, &quot;fun!&quot;]) |&gt; doc.to_string(80)\n// -&gt; [ &quot;Gleam&quot;, &quot;is&quot;, &quot;fun!&quot; ]\n\npretty_list([&quot;Gleam&quot;, &quot;is&quot;, &quot;fun!&quot;]) |&gt; doc.to_string(10)\n// -&gt;\n// [\n// &quot;Gleam&quot;,\n// &quot;is&quot;,\n// &quot;fun&quot;  \n// ]\n```\n\nWe&#39;re getting there! There&#39;s just a couple of things we need to fix before\ncalling it a day:\n\n- The list items should be indented by two spaces\n- When the list is rendered on multiple lines, the last item should have a\n  trailing comma\n- When the list is rendered on a single line we do not want the first and last\n  items to be separated with a space from the brackets\n\n## Nesting things\n\nTo solve the first problem, we need to add a some nesting whitespace whenever\na space is turned into a newline. In order to do so we can use the `doc.nest`\nfunction:\n\n```gleam\ndoc.nest(my_document, by: 2)\n```\n\nThis tells the pretty printer that, whenever it renders a space as a newline\nwhen formatting `my_document` it sould also indent it by two spaces.\nLet&#39;s get back to the `food_i_love` example and see how the rendering changes\nby adding a nesting of two spaces:\n\n```gleam\nfood_i_love\n|&gt; doc.nest(by: 2)\n|&gt; doc.to_string(20)\n\n// -&gt;\n// Food I love:\n//   lasagna\n//   ravioli\n//   pizza\n```\n\nNow we can add nesting to the `pretty_list` document:\n\n```gleam\npub fn pretty_list(list: List(String)) -&gt; Document {\n  let list_item_to_document = fn(item) { doc.from_string(&quot;\\&quot;&quot; &lt;&gt; item &lt;&gt; &quot;\\&quot;&quot;) }\n  let comma = doc.concat([doc.from_string(&quot;,&quot;), doc.space])\n  let open_square = doc.concat([doc.from_string(&quot;[&quot;), doc.space])\n  let close_square = doc.concat([doc.space, doc.from_string(&quot;]&quot;)])\n\n  list.map(list, list_item_to_document)\n  |&gt; doc.join(with: comma)\n  |&gt; doc.prepend(open_square)\n  |&gt; doc.nest(by: 2)\n  // ^-- every space before the final bracket will get indented\n  //     by two spaces if turned into a new line\n  |&gt; doc.append(close_square)\n  |&gt; doc.group\n}\n```\n\nNotice how the indentation applies to all the spaces _except_ the one that comes\nbefore the closed square bracket. This way, the closed square bracket will not\nbe indented and kept on the same nesting level as the corresponding open\nbracket:\n\n```gleam\npretty_list([&quot;Gleam&quot;, &quot;is&quot;, &quot;fun!&quot;]) |&gt; doc.to_string(10)\n// -&gt;\n// [\n//   &quot;Gleam&quot;,\n//   &quot;is&quot;,\n//   &quot;fun&quot;\n// ]\n```\n\n## Custom breaks\n\nThe last piece of work we need to do to get the desired look for a pretty\nprinted list is adding a trailing comma and removing whitespace before and\nafter the square brackets:\n\n```gleam\n// This:\n[&quot;Gleam&quot;, &quot;is&quot;, &quot;fun!&quot;]\n\n// not this:\n[ &quot;Gleam&quot;, &quot;is&quot;, &quot;fun!&quot; ]\n```\n\n```gleam\n// This:\n[\n  &quot;Gleam&quot;,\n  &quot;is&quot;,\n  &quot;fun!&quot;,\n]\n\n// not this:\n[\n  &quot;Gleam&quot;,\n  &quot;is&quot;,\n  &quot;fun!&quot;\n]\n```\n\nBoth problema can be solved by introducing a new powerful document constructor:\n`doc.break`.\n\nUp until now `doc.space` was the only document that the pretty printer could\nturn into a newline when splitting a group. Turns out that `doc.break` is a\nmore general version of `doc.space`. Here is how it can be called:\n\n```gleam\ndoc.break(&quot;unbroken&quot;, &quot;broken&quot;)\n```\n\n- It&#39;s first argument is the string that will be displayed if the pretty\n  printer decides not to break the document\n- It&#39;s second argument is the string that will be displayed if the pretty\n  printer decides to break the document _before_ inserting a newline\n\n### The pretty printing algorithm\n\nThis description of `doc.break` may sound a bit hand wavy so, to clear things\nup, let&#39;s have a final look at how the formatting of documents actually works:\n\n- The pretty printer always tries to put a group onto a single line by treating\n  every `doc.break` as if it were its first argument\n- If the group exceeds the maximum line width then all its `break`s are rendered\n  as their second argument and immediately followed by a newline\n  - If the document is also nested, the pretty printer adds a nesting of the\n    given level after the newline\n- After breaking the outermost group, the pretty printer considers every nested\n group on its own, deciding once again if it can fit on the new line or not\n\nLet&#39;s look at an example:\n\n```gleam\nlet visible_space = doc.break(&quot;•&quot;, &quot;↩&quot;)\nlet doc =\n  [&quot;Gleam&quot;, &quot;is&quot;, &quot;fun!&quot;]\n  |&gt; doc.join(with: visible_space)\n  |&gt; dog.group\n\ndoc |&gt; doc.to_string(80)\n// -&gt; Gleam•is•fun!\n\ndoc |&gt; doc.to_string(10)\n// -&gt;\n// Gleam↩\n// is↩\n// fun\n```\n\nAs you can see, when the pretty printer decides to split the document, all of\nits breaks are displayed as `&quot;↩&quot;`; if it is not broken they are displayed as\n`&quot;•&quot;`.\n\n&gt; `doc.space` can actually be defined in terms of `doc.break` (and that&#39;s how\n&gt; it is defined in glam!) as `doc.break(&quot; &quot;, &quot;&quot;)`:\n&gt;\n&gt; - if the group is not broken we render a single whitespace\n&gt; - if the group gets broken we do not render anything and let the pretty\n&gt;   printer do its magic and add a newline after each space\n\n### Back to lists one last time\n\nThanks to break we can solve both our remaining problems: first of all we want\nto disaply a trailing comma after the last element _only if_ the list is split\nonto multiple lines:\n\n```gleam\nlet trailing_comma = break(&quot;&quot;, &quot;,&quot;)\n```\n\nThis comma will only be displayed if the group it belongs to is broken by the\npretty printer:\n\n```gleam\npub fn pretty_list(list: List(String)) -&gt; Document {\n  let list_item_to_document = fn(item) { doc.from_string(&quot;\\&quot;&quot; &lt;&gt; item &lt;&gt; &quot;\\&quot;&quot;) }\n  let comma = doc.concat([doc.from_string(&quot;,&quot;), doc.space])\n  let open_square = doc.concat([doc.from_string(&quot;[&quot;), doc.space])\n  let trailing_comma = break(&quot;&quot;, &quot;,&quot;)\n  let close_square = doc.concat([trailing_comma, doc.from_string(&quot;]&quot;)])\n  // ^-- we add the trailing comma right before the closing bracket\n\n  list.map(list, list_item_to_document)\n  |&gt; doc.join(with: comma)\n  |&gt; doc.prepend(open_square)\n  |&gt; doc.nest(by: 2)\n  |&gt; doc.append(close_square)\n  |&gt; doc.group\n}\n```\n\nAs a finishing touch we need to remove the whitespace that gets printed after\nthe open bracket. The perfect fit for this job is once again `doc.break`:\n\n```gleam\nlet soft_break = doc.break(&quot;&quot;, &quot;&quot;)\n```\n\n`soft_break` will never be rendered, it&#39;s only purpose it to act as a possible\nbreaking point for the pretty printer to split a group on multiple lines.\n\nSince `soft_break` can be useful in a lot of different settings, the glam\npackage already exposes it as a constant so that you do not have to redefine it\nevery time you need it:\n\n```gleam\npub fn pretty_list(list: List(String)) -&gt; Document {\n  let list_item_to_document = fn(item) { doc.from_string(&quot;\\&quot;&quot; &lt;&gt; item &lt;&gt; &quot;\\&quot;&quot;) }\n  let comma = doc.concat([doc.from_string(&quot;,&quot;), doc.space])\n  let open_square = doc.concat([doc.from_string(&quot;[&quot;), doc.soft_break])\n  // ^-- so that there won&#39;t be a whitespace after the open bracket\n  //     but the pretty printer can still split it\n  let trailing_comma = doc.break(&quot;&quot;, &quot;,&quot;)\n  let close_square = doc.concat([trailing_comma, doc.from_string(&quot;]&quot;)])\n\n  list.map(list, list_item_to_document)\n  |&gt; doc.join(with: comma)\n  |&gt; doc.prepend(open_square)\n  |&gt; doc.nest(by: 2)\n  |&gt; doc.append(close_square)\n  |&gt; doc.group\n}\n```\n\nAnd that&#39;s it, the pretty printer is complete!\n\n## Recap\n\nPhew! That was quite a long tutorial so give yourself a pat in the back and\nenjoy the beautifully pretty printed lists.\n\nWe covered most of the glam API and got an understanding of how each piece can\ninfluence the pretty printing process:\n\n- `doc.from_string` to generate documents from strings\n- `doc.concat` to join together bunch of documents\n- `doc.group` and `doc.space` to allow the pretty printer to split documents\n- `doc.nest` to nest documents when the pretty printer decides to split them\n- `doc.break` to conditionally render strings depending if their group gets\n  splitted or not\n\nWe&#39;ve also used some utility methods like:\n\n- `doc.join` which is equivalent to interspersing and then using `doc.concat`\n- `doc.prepend` to prepend a document to an existing one\n- `doc.append` to append a document to an existing one\n\nIf you want more examples you can have a look at the\n[`examples`](https://github.com/giacomocavalieri/glam/tree/main/src/examples)\nfolder in the glam repo.\n\n---\n\nThanks for following along! If you spot any spelling mistake, wrong code\nsnippets or there are parts that are not clear, please let me know!\n\nYour feedback really means a lot to me so don&#39;t be scared to reach out!\nMy Twitter handle is [@giacomo_cava](https://twitter.com/giacomo_cava) and I\nusually hang around in the [Gleam Discord server](https://discord.gg/Fm8Pwmy),\nso you can ping me there, I&#39;m `jak11` (also the Gleam Discord is full of\namazing people so give it a try anyway ✨)\n","url":"01_gleam_lists.html"},{"doc":"glam/doc","title":"Document","content":"pub opaque type Document\n A document that can be pretty printed with `to_string`.\n \n\nSynonyms:\nglam/doc.Document\nglam/doc Document","url":"glam/doc.html#Document"},{"doc":"glam/doc","title":"empty","content":"pub const empty: Document = Concat([])\n An empty document that is printed as the empty string.\n \n ## Examples\n \n \n empty |&gt; to_string(80)\n // -&gt; &quot;&quot;\n \n \nSynonyms:\nglam/doc.empty\nglam/doc empty","url":"glam/doc.html#empty"},{"doc":"glam/doc","title":"line","content":"pub const line: Document = Line(1)\n A document that is always printed as a single newline.\n \n ## Examples\n \n \n line |&gt; to_string(80)\n // -&gt; &quot;\\n&quot;\n \nSynonyms:\nglam/doc.line\nglam/doc line","url":"glam/doc.html#line"},{"doc":"glam/doc","title":"soft_break","content":"pub const soft_break: Document = Break(&quot;&quot;, &quot;&quot;)\n A document that is always rendered as an empty string but can act as a\n breaking point for the pretty printer.\n \n This is a shorthand for `break(&quot;&quot;, &quot;&quot;)`.\n \n ## Examples\n \n \n let doc = [from_string(&quot;soft&quot;), soft_break, from_string(&quot;break&quot;)]\n \n doc |&gt; to_string(80)\n // -&gt; &quot;softbreak&quot;\n \n doc |&gt; to_string(5)\n // -&gt; &quot;soft\n // break&quot;\n \n \nSynonyms:\nglam/doc.soft_break\nglam/doc soft_break","url":"glam/doc.html#soft_break"},{"doc":"glam/doc","title":"space","content":"pub const space: Document = Break(&quot; &quot;, &quot;&quot;)\n A document that is rendered as a single whitespace `&quot; &quot;` but can be broken\n by the pretty printer into newlines instead.\n \n This is a shorthand for `break(&quot; &quot;, &quot;&quot;)`.\n \n ## Examples\n \n \n let doc =\n   [&quot;pretty&quot;, &quot;printed&quot;]\n   |&gt; list.map(from_string)\n   |&gt; join(with: space)\n \n doc |&gt; to_string(80)\n // -&gt; &quot;pretty printed&quot;\n \n doc |&gt; to_string(10)\n // -&gt; &quot;pretty\n // printed&quot;\n \n \nSynonyms:\nglam/doc.space\nglam/doc space","url":"glam/doc.html#space"},{"doc":"glam/doc","title":"append","content":"pub fn append(to first: Document, doc second: Document) -&gt; Document\n Joins a document into the end of another.\n \n ## Examples\n \n \n from_string(&quot;pretty&quot;)\n |&gt; append(from_string(&quot; printer&quot;))\n |&gt; to_string(80)\n // -&gt; &quot;pretty printer&quot;\n \n \nSynonyms:\nglam/doc.append\nglam/doc append","url":"glam/doc.html#append"},{"doc":"glam/doc","title":"append_docs","content":"pub fn append_docs(first: Document, docs: List(Document)) -&gt; Document\n Joins multiple documents into the end of another.\n \n This is a shorthand for `append(to: first, doc: concat(docs))`.\n \n ## Examples\n \n \n from_string(&quot;pretty&quot;)\n |&gt; append_docs([\n   from_string(&quot;printing&quot;),\n   space,\n   from_string(&quot;rocks!&quot;),\n ])\n |&gt; to_string(80) \n // -&gt; &quot;pretty printing rocks!&quot;\n \n \nSynonyms:\nglam/doc.append_docs\nglam/doc append_docs","url":"glam/doc.html#append_docs"},{"doc":"glam/doc","title":"break","content":"pub fn break(unbroken: String, broken: String) -&gt; Document\n A document after which the pretty printer can insert a newline.\n A newline is added after a `break` document if the `group` it is part of\n could not be rendered on a single line.\n \n If the pretty printer decides to add a newline after `break` it will be\n rendered as its first argument, otherwise as its first argument.\n \n ## Examples\n \n \n let message =\n   [from_string(&quot;pretty&quot;), break(&quot;•&quot;, &quot;↩&quot;), from_string(&quot;printed&quot;)]\n   |&gt; concat\n   |&gt; group\n \n message |&gt; to_string(20)\n // -&gt; &quot;pretty•printed&quot;\n \n message |&gt; to_string(10)\n // -&gt; &quot;pretty↩\n // printed&quot;\n \n \nSynonyms:\nglam/doc.break\nglam/doc break","url":"glam/doc.html#break"},{"doc":"glam/doc","title":"concat","content":"pub fn concat(docs: List(Document)) -&gt; Document\n Joins a list of documents into a single document.\n \n The resulting pretty printed document would be the same as pretty printing\n each document separately and concatenating it together with `&lt;&gt;`:\n \n \n docs |&gt; concat |&gt; to_string(n) ==\n docs |&gt; list.map(to_string(n)) |&gt; string.concat\n \n \n ## Examples\n \n \n [&quot;pretty&quot;, &quot; &quot;, &quot;printed&quot;]\n |&gt; list.map(from_string)\n |&gt; concat\n |&gt; to_string(80)\n // -&gt; &quot;pretty printed&quot;\n \n \nSynonyms:\nglam/doc.concat\nglam/doc concat","url":"glam/doc.html#concat"},{"doc":"glam/doc","title":"concat_join","content":"pub fn concat_join(docs: List(Document), with separators: List(\n    Document,\n  )) -&gt; Document\n Joins a list of documents into a single one by inserting the given\n separators between each existing document.\n \n This is a shorthand for `join(docs, concat(separators))`.\n \n ## Examples\n \n \n [&quot;wow&quot;, &quot;so&quot;, &quot;many&quot;, &quot;commas&quot;]\n |&gt; list.map(from_string)\n |&gt; concat_join([from_string(&quot;,&quot;), space])\n |&gt; to_string(80)\n // -&gt; &quot;wow, so, many, commas&quot;\n \n \nSynonyms:\nglam/doc.concat_join\nglam/doc concat_join","url":"glam/doc.html#concat_join"},{"doc":"glam/doc","title":"force_break","content":"pub fn force_break(doc: Document) -&gt; Document\n Forces the pretty printer to break all the `break`s of the outermost\n document. This still has no effect on `group`s as the pretty printer will\n always try to put them on a single line before splitting them.\n \n ## Examples\n \n \n [from_string(&quot;pretty&quot;), break(&quot;•&quot;, &quot;↩&quot;), from_string(&quot;printed&quot;)]\n |&gt; concat\n |&gt; force_break\n |&gt; group\n |&gt; to_string(100)\n // -&gt; &quot;pretty↩\n // printed&quot;\n \n \nSynonyms:\nglam/doc.force_break\nglam/doc force_break","url":"glam/doc.html#force_break"},{"doc":"glam/doc","title":"from_string","content":"pub fn from_string(string: String) -&gt; Document\n Turns a string into a document.\n \n ## Examples\n \n \n &quot;doc&quot; |&gt; from_string |&gt; to_string(80)\n // -&gt; &quot;doc&quot;\n \n \nSynonyms:\nglam/doc.from_string\nglam/doc from_string","url":"glam/doc.html#from_string"},{"doc":"glam/doc","title":"group","content":"pub fn group(doc: Document) -&gt; Document\n Allows the pretty printer to break the `break` documents inside the given\n group.\n \n When the pretty printer runs into a group it first tries to render it on a\n single line, displaying all the breaks as their first argument.\n If the group fits this is the final pretty printed result.\n \n However, if the group does not fit on a single line _all_ the `break`s\n inside that group are rendered as their second argument and immediately\n followed by a newline.\n \n Any nested group is considered on its own and may or may not be splitted,\n depending if it fits on a single line or not. So, even if the outermost\n group is broken, its nested groups may still end up on a single line.\n \n ## Examples\n \n \n let food =\n   [&quot;lasagna&quot;, &quot;ravioli&quot;, &quot;pizza&quot;]\n   |&gt; list.map(from_string) |&gt; join(with: space) |&gt; group\n let message = \n   [from_string(&quot;Food I love:&quot;), space, food] |&gt; concat |&gt; group\n \n message |&gt; to_string(80)\n // -&gt; &quot;Food I love: lasagna ravioli pizza&quot;\n \n message |&gt; to_string(30)\n // -&gt; &quot;Food I love:\n // lasagna ravioli pizza&quot;\n // ^-- After splitting the outer group, the inner one can fit\n //     on a single line so the pretty printer does not split it\n \n message |&gt; to_string(20)\n // &quot;Food I love:\n // lasagna\n // ravioli\n // pizza&quot;\n // ^-- Even after splitting the outer group, the inner one wouldn&#39;t\n //     fit on a single line, so the pretty printer splits that as well\n \n \nSynonyms:\nglam/doc.group\nglam/doc group","url":"glam/doc.html#group"},{"doc":"glam/doc","title":"join","content":"pub fn join(docs: List(Document), with separator: Document) -&gt; Document\n Joins a list of documents inserting the given separator between\n each existing document.\n \n ## Examples\n \n \n let message =\n   [&quot;Gleam&quot;, &quot;is&quot;, &quot;fun!&quot;]\n   |&gt; list.map(from_string)\n   |&gt; join(with: space)\n \n message |&gt; to_string(80)\n // -&gt; &quot;Gleam is fun!&quot;\n \n \nSynonyms:\nglam/doc.join\nglam/doc join","url":"glam/doc.html#join"},{"doc":"glam/doc","title":"lines","content":"pub fn lines(size: Int) -&gt; Document\n A document that is always printed as a series of consecutive newlines.\n \n ## Examples\n \n \n lines(3) |&gt; to_string(80)\n // -&gt; &quot;\\n\\n\\n&quot;\n \n \nSynonyms:\nglam/doc.lines\nglam/doc lines","url":"glam/doc.html#lines"},{"doc":"glam/doc","title":"nest","content":"pub fn nest(doc: Document, by indentation: Int) -&gt; Document\n Increases the nesting level of a document by the given amount.\n \n When the pretty printer breaks a group by inserting a newline, it also adds\n a whitespace padding equal to its nesting level.\n \n ## Examples\n \n \n let one = [space, from_string(&quot;one&quot;)] |&gt; concat |&gt; nest(by: 1)\n let two = [space, from_string(&quot;two&quot;)] |&gt; concat |&gt; nest(by: 2)\n let three = [space, from_string(&quot;three&quot;)] |&gt; concat |&gt; nest(by: 3)\n let list = [from_string(&quot;list:&quot;), one, two, three] |&gt; concat |&gt; group\n \n list |&gt; to_string(10)\n // -&gt; &quot;list:\n //  one\n //   two\n //    three&quot;\n \n\nSynonyms:\nglam/doc.nest\nglam/doc nest","url":"glam/doc.html#nest"},{"doc":"glam/doc","title":"prepend","content":"pub fn prepend(to first: Document, doc second: Document) -&gt; Document\n Prefixes a document to another one.\n \n ## Examples\n \n \n from_string(&quot;printed!&quot;)\n |&gt; prepend(from_string(&quot;pretty &quot;))\n |&gt; to_string(80)\n // -&gt; &quot;pretty printed!&quot;\n \n \nSynonyms:\nglam/doc.prepend\nglam/doc prepend","url":"glam/doc.html#prepend"},{"doc":"glam/doc","title":"prepend_docs","content":"pub fn prepend_docs(first: Document, docs: List(Document)) -&gt; Document\n Prefixes multiple documents to another one.\n \n This is a shorthand for `prepend(to: first, doc: concat(docs))`.\n \n ## Examples\n \n \n from_string(&quot;fun!&quot;)\n |&gt; prepend_docs([from_string(&quot;Gleam &quot;), from_string(&quot;is &quot;)])\n |&gt; to_string(80)\n // -&gt; &quot;Gleam is fun!&quot;\n \n \nSynonyms:\nglam/doc.prepend_docs\nglam/doc prepend_docs","url":"glam/doc.html#prepend_docs"},{"doc":"glam/doc","title":"to_string","content":"pub fn to_string(doc: Document, width: Int) -&gt; String\n Turns a document into a pretty printed string.\n The pretty printed process can be thought as follows:\n - the pretty printer first tries to print every group on a single line\n - all the `break` documents are rendered as their first argument\n - if the string fits on the specified width this is the result\n - if the string does not fit on a single line the outermost group is split:\n   - all of its `break` documents are rendered as their second argument\n   - a newline is inserted after every `break`\n   - a padding of the given nesting level is added after every inserted\n     newline\n   - all inner groups are then considered on their own: the splitting of the\n     outermost group does not imply that the inner gruops will be split as\n     well\n \n ## Examples\n \n For some examples on how pretty printing works for each kind of document you\n can have a look at the package documentation.\n There&#39;s also a\n [step by step tutorial](https://hexdocs.pm/glam/01_gleam_lists.html)\n that will guide you through the implementation of a simple pretty printer,\n covering most of the glam api.\n \nSynonyms:\nglam/doc.to_string\nglam/doc to_string","url":"glam/doc.html#to_string"},{"doc":"glam/doc","title":"to_string_builder","content":"pub fn to_string_builder(doc: Document, width: Int) -&gt; StringBuilder\n This function works like `to_string` but returns a `StringBuilder` instead\n of a `String`.\n \nSynonyms:\nglam/doc.to_string_builder\nglam/doc to_string_builder","url":"glam/doc.html#to_string_builder"},{"doc":"glam/doc","title":"glam/doc","content":"","url":"glam/doc.html"}]);